from typing import Union
from collections import deque
from threading import Lock
import time


class CallTracker:

    def __init__(
        self,
        *,
        chat_ttl_secs: Union[int, float, None] = None,
        user_ttl_secs: Union[int, float, None] = None,
        chat_user_ttl_secs: Union[int, float, None] = None
    ):
        self._chat_ttl_secs = chat_ttl_secs
        self._user_ttl_secs = user_ttl_secs
        self._chat_user_ttl_secs = chat_user_ttl_secs
        self._chat_call_times: dict[int, deque[float]] = {}
        self._user_call_times: dict[int, deque[float]] = {}
        self._chat_user_call_times: dict[int, dict[int, deque[float]]] = {}
        self._chat_call_time_lock = Lock()
        self._user_call_time_lock = Lock()
        self._chat_user_call_time_lock = Lock()

    def add_chat_call(self, chat_id: int) -> None:
        with self._chat_call_time_lock:
            self._chat_call_times.setdefault(chat_id, deque()).append(time.monotonic())

    def add_user_call(self, user_id: int) -> None:
        with self._user_call_time_lock:
            self._user_call_times.setdefault(user_id, deque()).append(time.monotonic())

    def add_chat_user_call(self, chat_id: int, user_id: int) -> None:
        with self._chat_user_call_time_lock:
            users = self._chat_user_call_times.setdefault(chat_id, {})
            call_times = users.setdefault(user_id, deque())
            call_times.append(time.monotonic())

    def get_chat_calls(self, chat_id: int) -> int:
        with self._chat_call_time_lock:
            chat_call_times = self._chat_call_times.get(chat_id)

            if chat_call_times is not None:
                if self._chat_ttl_secs is not None:
                    _remove_expired_call_times(chat_call_times, self._chat_ttl_secs)

                    if not chat_call_times:
                        del self._chat_call_times[chat_id]

                return len(chat_call_times)

            return 0

    def get_user_calls(self, user_id: int) -> int:
        with self._user_call_time_lock:
            user_call_times = self._user_call_times.get(user_id)

            if user_call_times is not None:
                if self._user_ttl_secs is not None:
                    _remove_expired_call_times(user_call_times, self._user_ttl_secs)

                    if not user_call_times:
                        del self._user_call_times[user_id]

                return len(user_call_times)

            return 0

    def get_chat_user_calls(self, chat_id: int, user_id: int) -> int:
        with self._chat_user_call_time_lock:
            chat_user_calls = self._chat_user_call_times.get(chat_id, {}).get(user_id)

            if chat_user_calls is not None:
                if self._chat_user_ttl_secs is not None:
                    _remove_expired_call_times(chat_user_calls, self._chat_user_ttl_secs)

                    if not chat_user_calls:
                        del self._chat_user_call_times[chat_id][user_id]

                        if not self._chat_user_call_times[chat_id]:
                            del self._chat_call_times[chat_id]

                return len(chat_user_calls)

            return 0


def _remove_expired_call_times(call_times: deque[float], ttl_secs: Union[int, float]) -> None:
    current_time = time.monotonic()
    expired_calls = 0

    for i in call_times:
        if (current_time - i) > ttl_secs:
            expired_calls += 1
        else:
            break

    for _ in range(expired_calls):
        call_times.popleft()



from typing import Union

from telebox.dispatcher.filters.event_filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.telegram_bot.types.types.message import Message
from telebox.telegram_bot.consts import message_entity_types


class CashtagFilter(AbstractEventFilter):

    def __init__(self, *cashtags: str):
        self._cashtags = set()

        for i in cashtags:
            if not i.startswith("$"):
                i = f"${i}"

            self._cashtags.add(i.upper())

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP
        }

    def get_value(self, event: Union[Message, MediaGroup]) -> set[str]:
        cashtags = set()

        if isinstance(event, MediaGroup):
            for i in event:
                _gather_cashtags(i, cashtags)
        else:
            _gather_cashtags(event, cashtags)

        return cashtags

    def check_value(self, value: set[str]) -> bool:
        return any(i in self._cashtags for i in value) if self._cashtags else bool(value)


def _gather_cashtags(message: Message, cashtags: set[str]) -> None:
    for i in message.get_entities():
        if i.type == message_entity_types.CASHTAG:
            cashtags.add(message.get_entity_text(i))



from typing import Optional

from telebox.dispatcher.filters.event_filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.telegram_bot.types.types.callback_query import CallbackQuery


class SimpleCallbackDataFilter(AbstractEventFilter):

    def __init__(self, *data: str):
        self._data = data

    def get_event_types(self) -> set[EventType]:
        return {EventType.CALLBACK_QUERY}

    def get_value(self, event: CallbackQuery) -> Optional[str]:
        return event.data

    def check_value(self, value: Optional[str]) -> bool:
        return value in self._data if self._data else value is not None



from typing import Union

from telebox.dispatcher.filters.event_filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.telegram_bot.types.types.message import Message
from telebox.telegram_bot.consts import message_entity_types


class HashtagFilter(AbstractEventFilter):

    def __init__(self, *hashtags: str, ignore_case: bool = True):
        self._hashtags = set()

        for i in hashtags:
            if not i.startswith("#"):
                i = f"#{i}"

            if ignore_case:
                i = i.lower()

            self._hashtags.add(i)

        self._ignore_case = ignore_case

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP
        }

    def get_value(self, event: Union[Message, MediaGroup]) -> set[str]:
        hashtags = set()

        if isinstance(event, MediaGroup):
            for i in event:
                _gather_hashtags(i, hashtags)
        else:
            _gather_hashtags(event, hashtags)

        return hashtags

    def check_value(self, value: set[str]) -> bool:
        if not self._hashtags:
            return bool(value)

        for i in value:
            if self._ignore_case:
                i = i.lower()

            if i in self._hashtags:
                return True

        return False


def _gather_hashtags(message: Message, hashtags: set[str]) -> None:
    for i in message.get_entities():
        if i.type == message_entity_types.HASHTAG:
            hashtags.add(message.get_entity_text(i))



from abc import ABC, abstractmethod


class AbstractCallbackDataBuilder(ABC):

    @abstractmethod
    def build(self, key, value=None) -> str:
        pass

    @abstractmethod
    def parse(self, data: str):
        pass



from typing import Union

from telebox.dispatcher.filters.event_filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.telegram_bot.types.types.message import Message


class ReplyMessageFilter(AbstractEventFilter):

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP
        }

    def get_value(self, event: Union[Message, MediaGroup]) -> bool:
        return event.is_reply

    def check_value(self, value: bool) -> bool:
        return value



from dataclasses import dataclass
from typing import Optional

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.consts import inline_query_result_types
from telebox.telegram_bot.types.types.inline_keyboard_markup import InlineKeyboardMarkup
from telebox.telegram_bot.types.types.input_message_content import InputMessageContent


@dataclass(unsafe_hash=True)
class InlineQueryResultArticle(Type):
    id: str
    title: str
    input_message_content: InputMessageContent
    reply_markup: Optional[InlineKeyboardMarkup] = None
    url: Optional[str] = None
    hide_url: Optional[bool] = None
    description: Optional[str] = None
    thumb_url: Optional[str] = None
    thumb_width: Optional[int] = None
    thumb_height: Optional[int] = None
    type: str = inline_query_result_types.ARTICLE



from __future__ import annotations
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from telebox.dispatcher.typing import Event
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.telegram_bot.types.types.message import Message
from telebox.telegram_bot.types.types.inline_query import InlineQuery
from telebox.telegram_bot.types.types.chosen_inline_result import ChosenInlineResult
from telebox.telegram_bot.types.types.callback_query import CallbackQuery
from telebox.telegram_bot.types.types.shipping_query import ShippingQuery
from telebox.telegram_bot.types.types.pre_checkout_query import PreCheckoutQuery
from telebox.telegram_bot.types.types.poll import Poll
from telebox.telegram_bot.types.types.poll_answer import PollAnswer
from telebox.telegram_bot.types.types.chat_member_updated import ChatMemberUpdated
from telebox.telegram_bot.types.types.chat_join_request import ChatJoinRequest


TYPES_WITH_CHAT_ID = frozenset({
    Message,
    MediaGroup,
    CallbackQuery,
    ChatMemberUpdated,
    ChatJoinRequest
})
TYPES_WITH_USER_ID = frozenset({
    Message,
    MediaGroup,
    InlineQuery,
    ChosenInlineResult,
    CallbackQuery,
    ShippingQuery,
    PreCheckoutQuery,
    Poll,
    PollAnswer,
    ChatMemberUpdated,
    ChatJoinRequest
})
TYPES_WITH_SENDER_CHAT_ID = frozenset({
    Message,
    MediaGroup
})
TYPES_WITH_MESSAGE_ID = frozenset({
    Message,
    CallbackQuery
})
TYPES_WITH_CALLBACK_QUERY_ID = frozenset({
    CallbackQuery
})
TYPES_WITH_INLINE_QUERY_ID = frozenset({
    InlineQuery
})
TYPES_WITH_SHIPPING_QUERY_ID = frozenset({
    ShippingQuery
})
TYPES_PRE_CHECKOUT_QUERY_ID = frozenset({
    PreCheckoutQuery
})


def get_event_chat_id(event: Event) -> Optional[int]:
    return event.chat_id if type(event) in TYPES_WITH_CHAT_ID else None


def get_event_user_id(event: Event) -> Optional[int]:
    return event.user_id if type(event) in TYPES_WITH_USER_ID else None


def get_event_sender_chat_id(event: Event) -> Optional[int]:
    return event.sender_chat_id if type(event) in TYPES_WITH_SENDER_CHAT_ID else None


def get_event_message_id(event: Event) -> Optional[int]:
    return event.message_id if type(event) in TYPES_WITH_MESSAGE_ID else None


def get_event_callback_query_id(event: Event) -> Optional[str]:
    return event.id if type(event) in TYPES_WITH_CALLBACK_QUERY_ID else None


def get_event_inline_query_id(event: Event) -> Optional[str]:
    return event.id if type(event) in TYPES_WITH_INLINE_QUERY_ID else None


def get_event_shipping_query_id(event: Event) -> Optional[str]:
    return event.id if type(event) in TYPES_WITH_SHIPPING_QUERY_ID else None


def get_event_pre_checkout_query_id(event: Event) -> Optional[str]:
    return event.id if type(event) in TYPES_PRE_CHECKOUT_QUERY_ID else None



from typing import Union

from telebox.telegram_bot.types.types.message import Message
from telebox.telegram_bot.types.types.inline_query import InlineQuery
from telebox.telegram_bot.types.types.chosen_inline_result import ChosenInlineResult
from telebox.telegram_bot.types.types.callback_query import CallbackQuery
from telebox.telegram_bot.types.types.shipping_query import ShippingQuery
from telebox.telegram_bot.types.types.pre_checkout_query import PreCheckoutQuery
from telebox.telegram_bot.types.types.poll import Poll
from telebox.telegram_bot.types.types.poll_answer import PollAnswer
from telebox.telegram_bot.types.types.chat_member_updated import ChatMemberUpdated
from telebox.telegram_bot.types.types.chat_join_request import ChatJoinRequest
from telebox.dispatcher.utils.media_group import MediaGroup


Event = Union[
    Message,
    MediaGroup,
    InlineQuery,
    ChosenInlineResult,
    CallbackQuery,
    ShippingQuery,
    PreCheckoutQuery,
    Poll,
    PollAnswer,
    ChatMemberUpdated,
    ChatJoinRequest
]



from .type import Type
from .types import (
    Update,
    WebhookInfo,
    User,
    Chat,
    Message,
    MessageId,
    MessageEntity,
    PhotoSize,
    Animation,
    Audio,
    Document,
    Video,
    VideoNote,
    Voice,
    Contact,
    Dice,
    PollOption,
    PollAnswer,
    Poll,
    Location,
    Venue,
    WebAppData,
    ProximityAlertTriggered,
    MessageAutoDeleteTimerChanged,
    ForumTopicCreated,
    ForumTopicClosed,
    ForumTopicReopened,
    VideoChatScheduled,
    VideoChatStarted,
    VideoChatEnded,
    VideoChatParticipantsInvited,
    UserProfilePhotos,
    File,
    WebAppInfo,
    ReplyKeyboardMarkup,
    KeyboardButton,
    KeyboardButtonPollType,
    ReplyKeyboardRemove,
    InlineKeyboardMarkup,
    InlineKeyboardButton,
    LoginUrl,
    CallbackQuery,
    ForceReply,
    ChatPhoto,
    ChatInviteLink,
    ChatAdministratorRights,
    ChatMember,
    ChatMemberOwner,
    ChatMemberAdministrator,
    ChatMemberMember,
    ChatMemberRestricted,
    ChatMemberLeft,
    ChatMemberBanned,
    ChatMemberUpdated,
    ChatJoinRequest,
    ChatPermissions,
    ChatLocation,
    ForumTopic,
    BotCommand,
    BotCommandScope,
    BotCommandScopeDefault,
    BotCommandScopeAllPrivateChats,
    BotCommandScopeAllGroupChats,
    BotCommandScopeAllChatAdministrators,
    BotCommandScopeChat,
    BotCommandScopeChatAdministrators,
    BotCommandScopeChatMember,
    MenuButton,
    MenuButtonCommands,
    MenuButtonWebApp,
    MenuButtonDefault,
    ResponseParameters,
    InputMedia,
    InputMediaPhoto,
    InputMediaVideo,
    InputMediaAnimation,
    InputMediaAudio,
    InputMediaDocument,
    InputFile,
    Sticker,
    StickerSet,
    MaskPosition,
    InlineQuery,
    InlineQueryResult,
    InlineQueryResultArticle,
    InlineQueryResultPhoto,
    InlineQueryResultGif,
    InlineQueryResultMpeg4Gif,
    InlineQueryResultVideo,
    InlineQueryResultAudio,
    InlineQueryResultVoice,
    InlineQueryResultDocument,
    InlineQueryResultLocation,
    InlineQueryResultVenue,
    InlineQueryResultContact,
    InlineQueryResultGame,
    InlineQueryResultCachedPhoto,
    InlineQueryResultCachedGif,
    InlineQueryResultCachedMpeg4Gif,
    InlineQueryResultCachedSticker,
    InlineQueryResultCachedDocument,
    InlineQueryResultCachedVideo,
    InlineQueryResultCachedVoice,
    InlineQueryResultCachedAudio,
    InputMessageContent,
    InputTextMessageContent,
    InputLocationMessageContent,
    InputVenueMessageContent,
    InputContactMessageContent,
    InputInvoiceMessageContent,
    ChosenInlineResult,
    SentWebAppMessage,
    LabeledPrice,
    Invoice,
    ShippingAddress,
    OrderInfo,
    ShippingOption,
    SuccessfulPayment,
    ShippingQuery,
    PreCheckoutQuery,
    PassportData,
    PassportFile,
    EncryptedPassportElement,
    EncryptedCredentials,
    PassportElementError,
    PassportElementErrorDataField,
    PassportElementErrorFrontSide,
    PassportElementErrorReverseSide,
    PassportElementErrorSelfie,
    PassportElementErrorFile,
    PassportElementErrorFiles,
    PassportElementErrorTranslationFile,
    PassportElementErrorTranslationFiles,
    PassportElementErrorUnspecified,
    Game,
    CallbackGame,
    GameHighScore
)


__all__ = [
    "Type",
    "Update",
    "WebhookInfo",
    "User",
    "Chat",
    "Message",
    "MessageId",
    "MessageEntity",
    "PhotoSize",
    "Animation",
    "Audio",
    "Document",
    "Video",
    "VideoNote",
    "Voice",
    "Contact",
    "Dice",
    "PollOption",
    "PollAnswer",
    "Poll",
    "Location",
    "Venue",
    "WebAppData",
    "ProximityAlertTriggered",
    "MessageAutoDeleteTimerChanged",
    "ForumTopicCreated",
    "ForumTopicClosed",
    "ForumTopicReopened",
    "VideoChatScheduled",
    "VideoChatStarted",
    "VideoChatEnded",
    "VideoChatParticipantsInvited",
    "UserProfilePhotos",
    "File",
    "WebAppInfo",
    "ReplyKeyboardMarkup",
    "KeyboardButton",
    "KeyboardButtonPollType",
    "ReplyKeyboardRemove",
    "InlineKeyboardMarkup",
    "InlineKeyboardButton",
    "LoginUrl",
    "CallbackQuery",
    "ForceReply",
    "ChatPhoto",
    "ChatInviteLink",
    "ChatAdministratorRights",
    "ChatMember",
    "ChatMemberOwner",
    "ChatMemberAdministrator",
    "ChatMemberMember",
    "ChatMemberRestricted",
    "ChatMemberLeft",
    "ChatMemberBanned",
    "ChatMemberUpdated",
    "ChatJoinRequest",
    "ChatPermissions",
    "ChatLocation",
    "ForumTopic",
    "BotCommand",
    "BotCommandScope",
    "BotCommandScopeDefault",
    "BotCommandScopeAllPrivateChats",
    "BotCommandScopeAllGroupChats",
    "BotCommandScopeAllChatAdministrators",
    "BotCommandScopeChat",
    "BotCommandScopeChatAdministrators",
    "BotCommandScopeChatMember",
    "MenuButton",
    "MenuButtonCommands",
    "MenuButtonWebApp",
    "MenuButtonDefault",
    "ResponseParameters",
    "InputMedia",
    "InputMediaPhoto",
    "InputMediaVideo",
    "InputMediaAnimation",
    "InputMediaAudio",
    "InputMediaDocument",
    "InputFile",
    "Sticker",
    "StickerSet",
    "MaskPosition",
    "InlineQuery",
    "InlineQueryResult",
    "InlineQueryResultArticle",
    "InlineQueryResultPhoto",
    "InlineQueryResultGif",
    "InlineQueryResultMpeg4Gif",
    "InlineQueryResultVideo",
    "InlineQueryResultAudio",
    "InlineQueryResultVoice",
    "InlineQueryResultDocument",
    "InlineQueryResultLocation",
    "InlineQueryResultVenue",
    "InlineQueryResultContact",
    "InlineQueryResultGame",
    "InlineQueryResultCachedPhoto",
    "InlineQueryResultCachedGif",
    "InlineQueryResultCachedMpeg4Gif",
    "InlineQueryResultCachedSticker",
    "InlineQueryResultCachedDocument",
    "InlineQueryResultCachedVideo",
    "InlineQueryResultCachedVoice",
    "InlineQueryResultCachedAudio",
    "InputMessageContent",
    "InputTextMessageContent",
    "InputLocationMessageContent",
    "InputVenueMessageContent",
    "InputContactMessageContent",
    "InputInvoiceMessageContent",
    "ChosenInlineResult",
    "SentWebAppMessage",
    "LabeledPrice",
    "Invoice",
    "ShippingAddress",
    "OrderInfo",
    "ShippingOption",
    "SuccessfulPayment",
    "ShippingQuery",
    "PreCheckoutQuery",
    "PassportData",
    "PassportFile",
    "EncryptedPassportElement",
    "EncryptedCredentials",
    "PassportElementError",
    "PassportElementErrorDataField",
    "PassportElementErrorFrontSide",
    "PassportElementErrorReverseSide",
    "PassportElementErrorSelfie",
    "PassportElementErrorFile",
    "PassportElementErrorFiles",
    "PassportElementErrorTranslationFile",
    "PassportElementErrorTranslationFiles",
    "PassportElementErrorUnspecified",
    "Game",
    "CallbackGame",
    "GameHighScore"
]



from typing import Optional
from threading import Lock

import ujson

from telebox.state_machine.storages.storage import AbstractStateStorage


class JSONStateStorage(AbstractStateStorage):

    def __init__(self, path: str):
        self._path = path
        self._lock = Lock()

    def save_states(
        self,
        states: list[str],
        *,
        chat_id: int,
        user_id: Optional[int] = None
    ) -> None:
        with self._lock:
            stored_states = self._load_stored_states()
            stored_states.setdefault(str(chat_id), {})[str(user_id)] = states

            with open(self._path, "w", encoding="UTF-8") as stream:
                ujson.dump(stored_states, stream)

    def load_states(self, *, chat_id: int, user_id: Optional[int] = None) -> list[str]:
        with self._lock:
            stored_states = self._load_stored_states()

            return stored_states.get(str(chat_id), {}).get(str(user_id), [])

    def _load_stored_states(self) -> dict[str, dict[str, list[str]]]:
        try:
            with open(self._path, encoding="UTF-8") as stream:
                return ujson.load(stream)
        except FileNotFoundError:
            return {}



from dataclasses import dataclass
from typing import Optional

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.types.types.location import Location


@dataclass(unsafe_hash=True)
class Venue(Type):
    location: Location
    title: str
    address: str
    foursquare_id: Optional[str] = None
    foursquare_type: Optional[str] = None
    google_place_id: Optional[str] = None
    google_place_type: Optional[str] = None



from dataclasses import dataclass

from telebox.telegram_bot.types.type import Type


@dataclass(unsafe_hash=True)
class MaskPosition(Type):
    point: str
    x_shift: float
    y_shift: float
    scale: float



from dataclasses import dataclass
from typing import Optional

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.consts import inline_query_result_types
from telebox.telegram_bot.types.types.message_entity import MessageEntity
from telebox.telegram_bot.types.types.inline_keyboard_markup import InlineKeyboardMarkup
from telebox.telegram_bot.types.types.input_message_content import InputMessageContent


@dataclass(unsafe_hash=True)
class InlineQueryResultVoice(Type):
    id: str
    voice_url: str
    title: str
    caption: Optional[str] = None
    parse_mode: Optional[str] = None
    caption_entities: Optional[list[MessageEntity]] = None
    voice_duration: Optional[int] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional[InputMessageContent] = None
    type: str = inline_query_result_types.VOICE



from dataclasses import dataclass
from typing import Optional, Literal

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.utils.users import get_full_name
from telebox.telegram_bot.utils.deep_links import get_username_link, get_user_link


@dataclass(unsafe_hash=True)
class User(Type):
    id: int
    is_bot: bool
    first_name: str
    last_name: Optional[str] = None
    username: Optional[str] = None
    language_code: Optional[str] = None
    is_premium: Optional[Literal[True]] = None
    added_to_attachment_menu: Optional[Literal[True]] = None
    can_join_groups: Optional[bool] = None
    can_read_all_group_messages: Optional[bool] = None
    supports_inline_queries: Optional[bool] = None

    @property
    def full_name(self) -> str:
        return get_full_name(self.first_name, self.last_name)

    @property
    def link(self) -> Optional[str]:
        if self.username is not None:
            return get_username_link(self.username)

    @property
    def mention_link(self) -> str:
        return get_user_link(self.id)



from dataclasses import dataclass, field
from typing import Optional

from telebox.telegram_bot.types.type import Type


@dataclass(unsafe_hash=True)
class InputFile(Type):
    content: bytes = field(repr=False)
    name: Optional[str] = None



from typing import Any, Callable

from telebox.dispatcher.typing import Event
from telebox.context.vars import event_context
from telebox.context.errors import InvalidEventError
from telebox.dispatcher.utils import events as event_utils


def get_event_chat_id() -> int:
    return _get_event_value(
        event_utils.get_event_chat_id,
        "chat identifier"
    )


def get_event_user_id() -> int:
    return _get_event_value(
        event_utils.get_event_user_id,
        "user identifier"
    )


def get_event_sender_chat_id() -> int:
    return _get_event_value(
        event_utils.get_event_sender_chat_id,
        "sender chat identifier"
    )


def get_event_message_id() -> int:
    return _get_event_value(
        event_utils.get_event_message_id,
        "message identifier"
    )


def get_event_callback_query_id() -> str:
    return _get_event_value(
        event_utils.get_event_callback_query_id,
        "callback query identifier"
    )


def get_event_inline_query_id() -> str:
    return _get_event_value(
        event_utils.get_event_inline_query_id,
        "inline query identifier"
    )


def get_event_shipping_query_id() -> str:
    return _get_event_value(
        event_utils.get_event_shipping_query_id,
        "shipping query identifier"
    )


def get_event_pre_checkout_query_id() -> str:
    return _get_event_value(
        event_utils.get_event_pre_checkout_query_id,
        "pre checkout query identifier"
    )


def _get_event_value(getter: Callable[[Event], Any], name: str) -> Any:
    event = event_context.get()
    value = getter(event)

    if value is None:
        raise InvalidEventError(f"No {name} in event" + " {event!r}!", event=event)

    return value



from typing import Callable

import cherrypy
import ujson

from telebox.telegram_bot.types.types.update import Update
from telebox.telegram_bot.serializer import Serializer


class ServerRoot:

    def __init__(self, update_processor: Callable[[Update], None]):
        self._serializer = Serializer()
        self._update_processor = update_processor

    @cherrypy.expose
    def index(self) -> str:
        content_length = cherrypy.request.headers.get("Content-Length")

        if content_length is None:
            raise cherrypy.HTTPError(403)

        data = ujson.loads(cherrypy.request.body.read(int(content_length)))
        update = self._serializer.get_object(data=data, class_=Update)
        self._update_processor(update)

        return str()



from typing import Union

from telebox.dispatcher.filters.event_filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.telegram_bot.types.types.message import Message


class TextFilter(AbstractEventFilter):

    def __init__(self, *texts: str, full_match: bool = True, ignore_case: bool = False):
        self._texts = {i.lower() for i in texts} if ignore_case else set(texts)
        self._full_match = full_match
        self._ignore_case = ignore_case

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP
        }

    def get_value(self, event: Union[Message, MediaGroup]) -> list[str]:
        if isinstance(event, MediaGroup):
            messages = event.messages
        else:
            messages = [event]

        texts = []

        for i in messages:
            text = i.get_text()

            if text is not None:
                texts.append(text)

        return texts

    def check_value(self, value: list[str]) -> bool:
        if not self._texts:
            return bool(value)

        for text in value:
            if self._ignore_case:
                text = text.lower()

            if self._full_match:
                if text in self._texts:
                    return True
            else:
                if any(i in text for i in self._texts):
                    return True

        return False



from dataclasses import dataclass

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.types.types.user import User


@dataclass(unsafe_hash=True)
class PollAnswer(Type):
    poll_id: str
    user: User
    option_ids: list[int]

    @property
    def user_id(self) -> int:
        return self.user.id



from typing import Union

from telebox.dispatcher.filters.event_filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.telegram_bot.types.types.message import Message
from telebox.telegram_bot.consts import message_entity_types


class MentionFilter(AbstractEventFilter):

    def __init__(self, *mentions: str):
        self._mentions = set()

        for i in mentions:
            if not i.startswith("@"):
                i = f"@{i}"

            self._mentions.add(i.lower())

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP
        }

    def get_value(self, event: Union[Message, MediaGroup]) -> set[str]:
        mentions = set()

        if isinstance(event, MediaGroup):
            for i in event:
                _gather_mentions(i, mentions)
        else:
            _gather_mentions(event, mentions)

        return mentions

    def check_value(self, value: set[str]) -> bool:
        return any(i in self._mentions for i in value) if self._mentions else bool(value)


def _gather_mentions(message: Message, mentions: set[str]) -> None:
    for i in message.get_entities():
        if i.type == message_entity_types.MENTION:
            mentions.add(message.get_entity_text(i).lower())



from typing import Union

from telebox.dispatcher.filters.event_filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.telegram_bot.types.types.message import Message


class ForwardedMessageFilter(AbstractEventFilter):

    def get_event_types(self) -> set[EventType]:
        return {EventType.MESSAGE, EventType.CHANNEL_POST, EventType.MEDIA_GROUP}

    def get_value(self, event: Union[Message, MediaGroup]) -> bool:
        return event.is_forwarded

    def check_value(self, value: bool) -> bool:
        return value



from dataclasses import dataclass
from typing import Optional

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.types.types.order_info import OrderInfo


@dataclass(unsafe_hash=True)
class SuccessfulPayment(Type):
    currency: str
    total_amount: int
    invoice_payload: str
    telegram_payment_charge_id: str
    provider_payment_charge_id: str
    shipping_option_id: Optional[str] = None
    order_info: Optional[OrderInfo] = None



from typing import Optional, Union

from telebox.dispatcher.filters.event_filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.telegram_bot.types.types.message import Message
from telebox.telegram_bot.consts import message_entity_types


class CommandFilter(AbstractEventFilter):

    def __init__(self, *commands: str, username: str, ignore_case: bool = True):
        self._commands = set()

        for i in commands:
            if not i.startswith("/"):
                i = f"/{i}"

            if ignore_case:
                i = i.lower()

            self._commands.update((i, f"{i}@{username.lower()}"))

        self._ignore_case = ignore_case

    def get_event_types(self) -> set[EventType]:
        return {
            EventType.MESSAGE,
            EventType.EDITED_MESSAGE,
            EventType.CHANNEL_POST,
            EventType.EDITED_CHANNEL_POST,
            EventType.MEDIA_GROUP
        }

    def get_value(self, event: Union[Message, MediaGroup]) -> Optional[str]:
        if isinstance(event, MediaGroup):
            for i in event:
                command = _get_command(i)

                if command is not None:
                    return command

        return _get_command(event)

    def check_value(self, value: Optional[str]) -> bool:
        if value is not None:
            if not self._commands:
                return True

            if self._ignore_case:
                value = value.lower()
            elif "@" in value:
                command, username = value.split("@", 1)
                value = f"{command}@{username.lower()}"

            return value in self._commands

        return False


def _get_command(message: Message) -> Optional[str]:
    for i in message.get_entities():
        if (i.type == message_entity_types.BOT_COMMAND) and (i.offset == 0):
            return message.get_entity_text(i)



from typing import Union

from telebox.telegram_bot.types.types.menu_button_commands import MenuButtonCommands
from telebox.telegram_bot.types.types.menu_button_web_app import MenuButtonWebApp
from telebox.telegram_bot.types.types.menu_button_default import MenuButtonDefault


MenuButton = Union[MenuButtonCommands,
                   MenuButtonWebApp,
                   MenuButtonDefault]



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Literal, Union, TYPE_CHECKING

from telebox.telegram_bot.consts import chat_types
from telebox.telegram_bot.utils.deep_links import get_message_public_link, get_message_private_link
from telebox.telegram_bot.enums.message_content_type import MessageContentType
from telebox.telegram_bot.errors import UnknownMessageContentError
from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.types.types.user import User
from telebox.telegram_bot.types.types.message_entity import MessageEntity
from telebox.telegram_bot.types.types.animation import Animation
from telebox.telegram_bot.types.types.audio import Audio
from telebox.telegram_bot.types.types.document import Document
from telebox.telegram_bot.types.types.photo_size import PhotoSize
from telebox.telegram_bot.types.types.sticker import Sticker
from telebox.telegram_bot.types.types.video import Video
from telebox.telegram_bot.types.types.video_note import VideoNote
from telebox.telegram_bot.types.types.voice import Voice
from telebox.telegram_bot.types.types.contact import Contact
from telebox.telegram_bot.types.types.dice import Dice
from telebox.telegram_bot.types.types.game import Game
from telebox.telegram_bot.types.types.poll import Poll
from telebox.telegram_bot.types.types.venue import Venue
from telebox.telegram_bot.types.types.location import Location
from telebox.telegram_bot.types.types.invoice import Invoice
from telebox.telegram_bot.types.types.successful_payment import SuccessfulPayment
from telebox.telegram_bot.types.types.passport_data import PassportData
from telebox.telegram_bot.types.types.proximity_alert_triggered import ProximityAlertTriggered
from telebox.telegram_bot.types.types.forum_topic_created import ForumTopicCreated
from telebox.telegram_bot.types.types.forum_topic_closed import ForumTopicClosed
from telebox.telegram_bot.types.types.forum_topic_reopened import ForumTopicReopened
from telebox.telegram_bot.types.types.video_chat_scheduled import VideoChatScheduled
from telebox.telegram_bot.types.types.video_chat_started import VideoChatStarted
from telebox.telegram_bot.types.types.video_chat_ended import VideoChatEnded
from telebox.telegram_bot.types.types.web_app_data import WebAppData
from telebox.telegram_bot.types.types.inline_keyboard_markup import InlineKeyboardMarkup
from telebox.telegram_bot.types.types.message_auto_delete_timer_changed import (
    MessageAutoDeleteTimerChanged
)
from telebox.telegram_bot.types.types.video_chat_participants_invited import (
    VideoChatParticipantsInvited
)
if TYPE_CHECKING:
    from telebox.telegram_bot.types.types.chat import Chat


@dataclass(unsafe_hash=True)
class Message(Type):
    message_id: int
    date: datetime
    chat: "Chat"
    message_thread_id: Optional[int] = None
    from_: Optional[User] = None
    sender_chat: Optional["Chat"] = None
    forward_from: Optional[User] = None
    forward_from_chat: Optional["Chat"] = None
    forward_from_message_id: Optional[int] = None
    forward_signature: Optional[str] = None
    forward_sender_name: Optional[str] = None
    forward_date: Optional[datetime] = None
    is_topic_message: Optional[Literal[True]] = None
    is_automatic_forward: Optional[Literal[True]] = None
    reply_to_message: Optional["Message"] = None
    via_bot: Optional[User] = None
    edit_date: Optional[datetime] = None
    has_protected_content: Optional[Literal[True]] = None
    media_group_id: Optional[str] = None
    author_signature: Optional[str] = None
    text: Optional[str] = None
    entities: Optional[list[MessageEntity]] = None
    animation: Optional[Animation] = None
    audio: Optional[Audio] = None
    document: Optional[Document] = None
    photo: Optional[list[PhotoSize]] = None
    sticker: Optional[Sticker] = None
    video: Optional[Video] = None
    video_note: Optional[VideoNote] = None
    voice: Optional[Voice] = None
    caption: Optional[str] = None
    caption_entities: Optional[list[MessageEntity]] = None
    contact: Optional[Contact] = None
    dice: Optional[Dice] = None
    game: Optional[Game] = None
    poll: Optional[Poll] = None
    venue: Optional[Venue] = None
    location: Optional[Location] = None
    new_chat_members: Optional[list[User]] = None
    left_chat_member: Optional[User] = None
    new_chat_title: Optional[str] = None
    new_chat_photo: Optional[list[PhotoSize]] = None
    delete_chat_photo: Optional[Literal[True]] = None
    group_chat_created: Optional[Literal[True]] = None
    supergroup_chat_created: Optional[Literal[True]] = None
    channel_chat_created: Optional[Literal[True]] = None
    message_auto_delete_timer_changed: Optional[MessageAutoDeleteTimerChanged] = None
    migrate_to_chat_id: Optional[int] = None
    migrate_from_chat_id: Optional[int] = None
    pinned_message: Optional["Message"] = None
    invoice: Optional[Invoice] = None
    successful_payment: Optional[SuccessfulPayment] = None
    connected_website: Optional[str] = None
    passport_data: Optional[PassportData] = None
    proximity_alert_triggered: Optional[ProximityAlertTriggered] = None
    forum_topic_created: Optional[ForumTopicCreated] = None
    forum_topic_closed: Optional[ForumTopicClosed] = None
    forum_topic_reopened: Optional[ForumTopicReopened] = None
    video_chat_scheduled: Optional[VideoChatScheduled] = None
    video_chat_started: Optional[VideoChatStarted] = None
    video_chat_ended: Optional[VideoChatEnded] = None
    video_chat_participants_invited: Optional[VideoChatParticipantsInvited] = None
    web_app_data: Optional[WebAppData] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None

    @property
    def content(self) -> tuple["MessageContent", MessageContentType]:
        if self.text is not None:
            return self.text, MessageContentType.TEXT
        elif self.animation is not None:
            return self.animation, MessageContentType.ANIMATION
        elif self.audio is not None:
            return self.audio, MessageContentType.AUDIO
        elif self.document is not None:
            return self.document, MessageContentType.DOCUMENT
        elif self.photo is not None:
            return self.photo, MessageContentType.PHOTO
        elif self.sticker is not None:
            return self.sticker, MessageContentType.STICKER
        elif self.video is not None:
            return self.video, MessageContentType.VIDEO
        elif self.video_note is not None:
            return self.video_note, MessageContentType.VIDEO_NOTE
        elif self.voice is not None:
            return self.voice, MessageContentType.VOICE
        elif self.contact is not None:
            return self.contact, MessageContentType.CONTACT
        elif self.dice is not None:
            return self.dice, MessageContentType.DICE
        elif self.game is not None:
            return self.game, MessageContentType.GAME
        elif self.poll is not None:
            return self.poll, MessageContentType.POLL
        elif self.venue is not None:
            return self.venue, MessageContentType.VENUE
        elif self.location is not None:
            return self.location, MessageContentType.LOCATION
        elif self.new_chat_members is not None:
            return self.new_chat_members, MessageContentType.NEW_CHAT_MEMBERS
        elif self.left_chat_member is not None:
            return self.left_chat_member, MessageContentType.LEFT_CHAT_MEMBER
        elif self.new_chat_title is not None:
            return self.new_chat_title, MessageContentType.NEW_CHAT_TITLE
        elif self.new_chat_photo is not None:
            return self.new_chat_photo, MessageContentType.NEW_CHAT_PHOTO
        elif self.delete_chat_photo is not None:
            return self.delete_chat_photo, MessageContentType.DELETE_CHAT_PHOTO
        elif self.group_chat_created is not None:
            return self.group_chat_created, MessageContentType.GROUP_CHAT_CREATED
        elif self.supergroup_chat_created is not None:
            return self.supergroup_chat_created, MessageContentType.SUPERGROUP_CHAT_CREATED
        elif self.channel_chat_created is not None:
            return self.channel_chat_created, MessageContentType.CHANNEL_CHAT_CREATED
        elif self.message_auto_delete_timer_changed is not None:
            return (
                self.message_auto_delete_timer_changed,
                MessageContentType.MESSAGE_AUTO_DELETE_TIMER_CHANGED
            )
        elif self.migrate_to_chat_id is not None:
            return self.migrate_to_chat_id, MessageContentType.MIGRATE_TO_CHAT_ID
        elif self.migrate_from_chat_id is not None:
            return self.migrate_from_chat_id, MessageContentType.MIGRATE_FROM_CHAT_ID
        elif self.pinned_message is not None:
            return self.pinned_message, MessageContentType.PINNED_MESSAGE
        elif self.invoice is not None:
            return self.invoice, MessageContentType.INVOICE
        elif self.successful_payment is not None:
            return self.successful_payment, MessageContentType.SUCCESSFUL_PAYMENT
        elif self.connected_website is not None:
            return self.connected_website, MessageContentType.CONNECTED_WEBSITE
        elif self.passport_data is not None:
            return self.passport_data, MessageContentType.PASSPORT_DATA
        elif self.proximity_alert_triggered is not None:
            return self.proximity_alert_triggered, MessageContentType.PROXIMITY_ALERT_TRIGGERED
        elif self.forum_topic_created is not None:
            return self.forum_topic_created, MessageContentType.FORUM_TOPIC_CREATED
        elif self.forum_topic_closed is not None:
            return self.forum_topic_closed, MessageContentType.FORUM_TOPIC_CLOSED
        elif self.forum_topic_reopened is not None:
            return self.forum_topic_reopened, MessageContentType.FORUM_TOPIC_REOPENED
        elif self.video_chat_scheduled is not None:
            return self.video_chat_scheduled, MessageContentType.VIDEO_CHAT_SCHEDULED
        elif self.video_chat_started is not None:
            return self.video_chat_started, MessageContentType.VIDEO_CHAT_STARTED
        elif self.video_chat_ended is not None:
            return self.video_chat_ended, MessageContentType.VIDEO_CHAT_ENDED
        elif self.video_chat_participants_invited is not None:
            return (
                self.video_chat_participants_invited,
                MessageContentType.VIDEO_CHAT_PARTICIPANTS_INVITED
            )
        elif self.web_app_data is not None:
            return self.web_app_data, MessageContentType.WEB_APP_DATA

        raise UnknownMessageContentError("Unknown message type!")

    @property
    def chat_type(self) -> str:
        return self.chat.type

    @property
    def chat_id(self) -> int:
        return self.chat.id

    @property
    def sender_chat_id(self) -> Optional[int]:
        return self.sender_chat.id if self.sender_chat is not None else None

    @property
    def user_id(self) -> Optional[int]:
        return self.from_.id if self.from_ is not None else None

    @property
    def is_forwarded(self) -> bool:
        return self.forward_date is not None

    @property
    def is_reply(self) -> bool:
        return self.reply_to_message is not None

    @property
    def link(self) -> Optional[str]:
        if self.chat.type in {chat_types.CHANNEL, chat_types.SUPERGROUP}:
            if self.chat.username is not None:
                return get_message_public_link(self.chat.username, self.message_id)
            else:
                return get_message_private_link(self.chat.id, self.message_id)

    def get_text(self) -> Optional[str]:
        if self.text is not None:
            return self.text
        elif self.caption is not None:
            return self.caption

    def get_entity_text(self, entity: MessageEntity) -> Optional[str]:
        text = self.get_text()

        if text is not None:
            bytes_ = text.encode("UTF-16-LE")
            entity_bytes = bytes_[entity.offset * 2:entity.end_offset * 2]

            return entity_bytes.decode("UTF-16-LE")

    def get_entities(self) -> list[MessageEntity]:
        if self.entities is not None:
            return self.entities
        elif self.caption_entities is not None:
            return self.caption_entities

        return []


MessageContent = Union[
    str,
    Animation,
    Audio,
    Document,
    list[PhotoSize],
    Sticker,
    Video,
    VideoNote,
    Voice,
    Contact,
    Dice,
    Game,
    Poll,
    Venue,
    Location,
    list[User],
    User,
    Literal[True],
    MessageAutoDeleteTimerChanged,
    int,
    Message,
    Invoice,
    SuccessfulPayment,
    PassportData,
    ProximityAlertTriggered,
    ForumTopicCreated,
    ForumTopicClosed,
    ForumTopicReopened,
    VideoChatScheduled,
    VideoChatStarted,
    VideoChatEnded,
    VideoChatParticipantsInvited,
    WebAppData
]



from dataclasses import dataclass
from typing import Optional

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.types.types.photo_size import PhotoSize
from telebox.telegram_bot.types.types.message_entity import MessageEntity
from telebox.telegram_bot.types.types.animation import Animation


@dataclass(unsafe_hash=True)
class Game(Type):
    title: str
    description: str
    photo: list[PhotoSize]
    text: Optional[str] = None
    text_entities: Optional[list[MessageEntity]] = None
    animation: Optional[Animation] = None



from typing import Type, TypeVar, Any, Optional, Callable
from datetime import datetime as datetime_, timezone
import dataclasses

from dataclass_factory import Factory, Schema

from telebox.telegram_bot.types.types.chat_member_owner import ChatMemberOwner
from telebox.telegram_bot.types.types.chat_member_administrator import ChatMemberAdministrator
from telebox.telegram_bot.types.types.chat_member_member import ChatMemberMember
from telebox.telegram_bot.types.types.chat_member_restricted import ChatMemberRestricted
from telebox.telegram_bot.types.types.chat_member_left import ChatMemberLeft
from telebox.telegram_bot.types.types.chat_member_banned import ChatMemberBanned
from telebox.telegram_bot.types.types.bot_command_scope_default import (
    BotCommandScopeDefault
)
from telebox.telegram_bot.types.types.bot_command_scope_all_private_chats import (
    BotCommandScopeAllPrivateChats
)
from telebox.telegram_bot.types.types.bot_command_scope_all_group_chats import (
    BotCommandScopeAllGroupChats
)
from telebox.telegram_bot.types.types.bot_command_scope_all_chat_administrators import (
    BotCommandScopeAllChatAdministrators
)
from telebox.telegram_bot.types.types.bot_command_scope_chat import (
    BotCommandScopeChat
)
from telebox.telegram_bot.types.types.bot_command_scope_chat_administrators import (
    BotCommandScopeChatAdministrators
)
from telebox.telegram_bot.types.types.bot_command_scope_chat_member import (
    BotCommandScopeChatMember
)
from telebox.telegram_bot.types.types.menu_button_commands import MenuButtonCommands
from telebox.telegram_bot.types.types.menu_button_web_app import MenuButtonWebApp
from telebox.telegram_bot.types.types.menu_button_default import MenuButtonDefault
from telebox.telegram_bot.types.types.input_media_photo import InputMediaPhoto
from telebox.telegram_bot.types.types.input_media_video import InputMediaVideo
from telebox.telegram_bot.types.types.input_media_animation import InputMediaAnimation
from telebox.telegram_bot.types.types.input_media_audio import InputMediaAudio
from telebox.telegram_bot.types.types.input_media_document import InputMediaDocument
from telebox.telegram_bot.types.types.inline_query_result_article import InlineQueryResultArticle
from telebox.telegram_bot.types.types.inline_query_result_photo import InlineQueryResultPhoto
from telebox.telegram_bot.types.types.inline_query_result_gif import InlineQueryResultGif
from telebox.telegram_bot.types.types.inline_query_result_mpeg4_gif import InlineQueryResultMpeg4Gif
from telebox.telegram_bot.types.types.inline_query_result_video import InlineQueryResultVideo
from telebox.telegram_bot.types.types.inline_query_result_audio import InlineQueryResultAudio
from telebox.telegram_bot.types.types.inline_query_result_voice import InlineQueryResultVoice
from telebox.telegram_bot.types.types.inline_query_result_document import InlineQueryResultDocument
from telebox.telegram_bot.types.types.inline_query_result_location import InlineQueryResultLocation
from telebox.telegram_bot.types.types.inline_query_result_venue import InlineQueryResultVenue
from telebox.telegram_bot.types.types.inline_query_result_contact import InlineQueryResultContact
from telebox.telegram_bot.types.types.inline_query_result_game import InlineQueryResultGame
from telebox.telegram_bot.types.types.inline_query_result_cached_photo import (
    InlineQueryResultCachedPhoto
)
from telebox.telegram_bot.types.types.inline_query_result_cached_gif import (
    InlineQueryResultCachedGif
)
from telebox.telegram_bot.types.types.inline_query_result_cached_mpeg4_gif import (
    InlineQueryResultCachedMpeg4Gif
)
from telebox.telegram_bot.types.types.inline_query_result_cached_sticker import (
    InlineQueryResultCachedSticker
)
from telebox.telegram_bot.types.types.inline_query_result_cached_document import (
    InlineQueryResultCachedDocument
)
from telebox.telegram_bot.types.types.inline_query_result_cached_video import (
    InlineQueryResultCachedVideo
)
from telebox.telegram_bot.types.types.inline_query_result_cached_voice import (
    InlineQueryResultCachedVoice
)
from telebox.telegram_bot.types.types.inline_query_result_cached_audio import (
    InlineQueryResultCachedAudio
)
from telebox.telegram_bot.types.types.passport_element_error_data_field import (
    PassportElementErrorDataField
)
from telebox.telegram_bot.types.types.passport_element_error_front_side import (
    PassportElementErrorFrontSide
)
from telebox.telegram_bot.types.types.passport_element_error_reverse_side import (
    PassportElementErrorReverseSide
)
from telebox.telegram_bot.types.types.passport_element_error_selfie import (
    PassportElementErrorSelfie
)
from telebox.telegram_bot.types.types.passport_element_error_file import (
    PassportElementErrorFile
)
from telebox.telegram_bot.types.types.passport_element_error_files import (
    PassportElementErrorFiles
)
from telebox.telegram_bot.types.types.passport_element_error_translation_file import (
    PassportElementErrorTranslationFile
)
from telebox.telegram_bot.types.types.passport_element_error_translation_files import (
    PassportElementErrorTranslationFiles
)
from telebox.telegram_bot.types.types.passport_element_error_unspecified import (
    PassportElementErrorUnspecified
)


DataclassObject = TypeVar("DataclassObject")
_DEFAULT_POST_SERIALIZATION_CLASSES = (
    # ChatMember types
    ChatMemberOwner,
    ChatMemberAdministrator,
    ChatMemberAdministrator,
    ChatMemberMember,
    ChatMemberRestricted,
    ChatMemberLeft,
    ChatMemberBanned,

    # BotCommandScope types
    BotCommandScopeDefault,
    BotCommandScopeAllPrivateChats,
    BotCommandScopeAllGroupChats,
    BotCommandScopeAllChatAdministrators,
    BotCommandScopeChat,
    BotCommandScopeChatAdministrators,
    BotCommandScopeChatMember,

    # MenuButton types
    MenuButtonCommands,
    MenuButtonWebApp,
    MenuButtonDefault,

    # InputMedia types
    InputMediaPhoto,
    InputMediaVideo,
    InputMediaAnimation,
    InputMediaAudio,
    InputMediaDocument,

    # InlineQueryResult types
    InlineQueryResultArticle,
    InlineQueryResultPhoto,
    InlineQueryResultGif,
    InlineQueryResultMpeg4Gif,
    InlineQueryResultVideo,
    InlineQueryResultAudio,
    InlineQueryResultVoice,
    InlineQueryResultDocument,
    InlineQueryResultLocation,
    InlineQueryResultVenue,
    InlineQueryResultContact,
    InlineQueryResultGame,
    InlineQueryResultCachedPhoto,
    InlineQueryResultCachedGif,
    InlineQueryResultCachedMpeg4Gif,
    InlineQueryResultCachedSticker,
    InlineQueryResultCachedDocument,
    InlineQueryResultCachedVideo,
    InlineQueryResultCachedVoice,
    InlineQueryResultCachedAudio,

    # PassportElementError types
    PassportElementErrorDataField,
    PassportElementErrorFrontSide,
    PassportElementErrorReverseSide,
    PassportElementErrorSelfie,
    PassportElementErrorFile,
    PassportElementErrorFiles,
    PassportElementErrorTranslationFile,
    PassportElementErrorTranslationFiles,
    PassportElementErrorUnspecified
)


def convert_datetime_to_timestamp(datetime: Optional[datetime_]) -> int:
    return int(datetime.timestamp()) if datetime is not None else 0


def convert_timestamp_to_datetime(timestamp: int) -> Optional[datetime_]:
    return datetime_.fromtimestamp(timestamp, tz=timezone.utc) if timestamp else None


class Serializer:

    # noinspection PyMethodMayBeStatic
    def get_object(
        self,
        data: dict[str, Any],
        class_: Type[DataclassObject]
    ) -> DataclassObject:
        return _dataclass_factory.load(data, class_)

    # noinspection PyMethodMayBeStatic
    def get_data(self, object_: Any) -> dict[str, Any]:
        return _dataclass_factory.dump(object_)


def _get_dataclass_factory() -> Factory:
    schemas = {
        datetime_: Schema(
            serializer=convert_datetime_to_timestamp,
            parser=convert_timestamp_to_datetime
        )
    }

    for class_ in _DEFAULT_POST_SERIALIZATION_CLASSES:
        for field in dataclasses.fields(class_):
            if isinstance(field.default, str):
                schemas[class_] = Schema(
                    post_serialize=_get_post_default_serializer(field.name, field.default)
                )

                break

    return Factory(
        default_schema=Schema(omit_default=True),
        schemas=schemas
    )


def _get_post_default_serializer(
    key: str,
    value: Any
) -> Callable[[dict[str, Any]], dict[str, Any]]:
    def _serializer(data: dict[str, Any]) -> dict[str, Any]:
        data[key] = value

        return data

    return _serializer


_dataclass_factory = _get_dataclass_factory()



from dataclasses import dataclass

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.types.types.user import User


@dataclass(unsafe_hash=True)
class VideoChatParticipantsInvited(Type):
    users: list[User]



from dataclasses import dataclass
from datetime import datetime

from telebox.telegram_bot.types.type import Type


@dataclass(unsafe_hash=True)
class VideoChatScheduled(Type):
    start_date: datetime



from dataclasses import dataclass

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.consts import passport_element_error_sources


@dataclass(unsafe_hash=True)
class PassportElementErrorFiles(Type):
    type: str
    file_hashes: list[str]
    message: str
    source: str = passport_element_error_sources.FILES



from .none import ErrorNoneFilter
from .type import ErrorTypeFilter
from .text import ErrorTextFilter
from .reg_exp_text import ErrorRegExpTextFilter


__all__ = [
    "ErrorNoneFilter",
    "ErrorTypeFilter",
    "ErrorTextFilter",
    "ErrorRegExpTextFilter"
]



DEFAULT = "default"
ALL_PRIVATE_CHATS = "all_private_chats"
ALL_GROUP_CHATS = "all_group_chats"
ALL_CHAT_ADMINISTRATORS = "all_chat_administrators"
CHAT = "chat"
CHAT_ADMINISTRATORS = "chat_administrators"
CHAT_MEMBER = "chat_member"



import logging
from typing import Optional, Union, NoReturn
from dataclasses import dataclass
from threading import Thread, Lock
import time

import cherrypy

from telebox.telegram_bot.bot import TelegramBot
from telebox.telegram_bot.types.types.update import Update
from telebox.telegram_bot.types.types.message import Message
from telebox.telegram_bot.errors import UnknownUpdateContentError, UnknownMessageContentError
from telebox.dispatcher.typing import Event
from telebox.dispatcher.utils.media_group import MediaGroup
from telebox.dispatcher.utils.event_queue import EventQueue
from telebox.dispatcher.enums.event_type import EventType
from telebox.dispatcher.handlers.event import AbstractEventHandler
from telebox.dispatcher.handlers.error import AbstractErrorHandler
from telebox.dispatcher.filters.event_filter import AbstractEventBaseFilter
from telebox.dispatcher.filters.error_filter import AbstractErrorBaseFilter
from telebox.dispatcher.filters.event.none import NoneFilter
from telebox.dispatcher.filters.error.none import ErrorNoneFilter
from telebox.dispatcher.middlewares.middleware import Middleware
from telebox.dispatcher.utils.rate_limiter.rate_limiter import RateLimiter
from telebox.dispatcher.utils.rate_limiter.rate_limit import RateLimit
from telebox.dispatcher.utils.timed_container import TimedContainer
from telebox.dispatcher.utils.server_root import ServerRoot
from telebox.dispatcher.utils.router import Router
from telebox.dispatcher.errors import DispatcherError
from telebox.utils.thread_pool import ThreadPool
from telebox.utils.not_set import NotSet, NOT_SET
from telebox.dispatcher.utils.events import get_event_chat_id
from telebox.context.vars import (
    event_context,
    event_handler_context,
    error_handler_context
)


logger = logging.getLogger(__name__)
_none_filter = NoneFilter()
_error_none_filter = ErrorNoneFilter()
_MEDIA_GROUP_GATHERING_DELAY_SECS = 0.1


@dataclass
class EventHandler:
    handler: AbstractEventHandler
    filter: AbstractEventBaseFilter
    rate_limiter: Optional[RateLimiter] = None

    @property
    def rate_limit(self) -> Optional[RateLimit]:
        return self.rate_limiter.limit if self.rate_limiter is not None else None


@dataclass
class ErrorHandler:
    handler: AbstractErrorHandler
    filter: AbstractErrorBaseFilter


class Dispatcher:

    def __init__(
        self,
        bot: TelegramBot,
        *,
        drop_over_limit_events: bool = False,
        default_rate_limit: Optional[RateLimit] = None
    ):
        self._bot = bot
        self._drop_over_limit_events = drop_over_limit_events
        self._default_rate_limit = default_rate_limit
        self._polling_is_used = False
        self._server_is_used = False
        self._events = EventQueue()
        self._thread_pool: Optional[ThreadPool] = None
        self._event_handlers: dict[EventType, list[EventHandler]] = {i: [] for i in EventType}
        self._error_handlers: list[ErrorHandler] = []
        self._middlewares: list[Middleware] = []
        self.router = Router(self)
        self._media_group_gathering_thread: Optional[Thread] = None
        self._media_group_messages: dict[str, TimedContainer] = {}
        self._media_group_message_lock = Lock()

    def add_message_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.MESSAGE, filter_, rate_limit)

    def add_edited_message_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.EDITED_MESSAGE, filter_, rate_limit)

    def add_channel_post_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.CHANNEL_POST, filter_, rate_limit)

    def add_edited_channel_post_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.EDITED_CHANNEL_POST, filter_, rate_limit)

    def add_media_group_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.MEDIA_GROUP, filter_, rate_limit)

    def add_inline_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.INLINE_QUERY, filter_, rate_limit)

    def add_chosen_inline_result_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.CHOSEN_INLINE_RESULT, filter_, rate_limit)

    def add_callback_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.CALLBACK_QUERY, filter_, rate_limit)

    def add_shipping_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.SHIPPING_QUERY, filter_, rate_limit)

    def add_pre_checkout_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._add_event_handler(handler, EventType.PRE_CHECKOUT_QUERY, filter_, rate_limit)

    def add_poll_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._add_event_handler(handler, EventType.POLL, filter_, None)

    def add_poll_answer_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._add_event_handler(handler, EventType.POLL_ANSWER, filter_, None)

    def add_my_chat_member_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._add_event_handler(handler, EventType.MY_CHAT_MEMBER, filter_, None)

    def add_chat_member_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._add_event_handler(handler, EventType.CHAT_MEMBER, filter_, None)

    def add_chat_join_request_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._add_event_handler(handler, EventType.CHAT_JOIN_REQUEST, filter_, None)

    def add_error_handler(
        self,
        handler: AbstractErrorHandler,
        filter_: Optional[AbstractErrorBaseFilter] = None
    ) -> None:
        filter_ = _get_error_filter(filter_)

        self._error_handlers.append(
            ErrorHandler(
                handler=handler,
                filter=filter_
            )
        )

    def add_middleware(self, middleware: Middleware) -> None:
        self._middlewares.append(middleware)

    def check_message_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.MESSAGE, filter_, rate_limit)

    def check_edited_message_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.EDITED_MESSAGE, filter_, rate_limit)

    def check_channel_post_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.CHANNEL_POST, filter_, rate_limit)

    def check_edited_channel_post_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(
            handler,
            EventType.EDITED_CHANNEL_POST,
            filter_,
            rate_limit
        )

    def check_media_group_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.MEDIA_GROUP, filter_, rate_limit)

    def check_inline_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.INLINE_QUERY, filter_, rate_limit)

    def check_chosen_inline_result_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(
            handler,
            EventType.CHOSEN_INLINE_RESULT,
            filter_,
            rate_limit
        )

    def check_callback_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.CALLBACK_QUERY, filter_, rate_limit)

    def check_shipping_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(handler, EventType.SHIPPING_QUERY, filter_, rate_limit)

    def check_pre_checkout_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        return self._check_event_handler(
            handler,
            EventType.PRE_CHECKOUT_QUERY,
            filter_,
            rate_limit
        )

    def check_poll_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> bool:
        return self._check_event_handler(handler, EventType.POLL, filter_, None)

    def check_poll_answer_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> bool:
        return self._check_event_handler(handler, EventType.POLL_ANSWER, filter_, None)

    def check_my_chat_member_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> bool:
        return self._check_event_handler(handler, EventType.MY_CHAT_MEMBER, filter_, None)

    def check_chat_member_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> bool:
        return self._check_event_handler(handler, EventType.CHAT_MEMBER, filter_, None)

    def check_chat_join_request_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> bool:
        return self._check_event_handler(handler, EventType.CHAT_JOIN_REQUEST, filter_, None)

    def check_error_handler(
        self,
        handler: AbstractErrorHandler,
        filter_: Optional[AbstractErrorBaseFilter] = None
    ) -> bool:
        filter_ = _get_error_filter(filter_)

        for i in self._error_handlers:
            if (i.handler is handler) and (i.filter is filter_):
                return True

        return False

    def check_middleware(self, middleware: Middleware) -> bool:
        for i in self._middlewares:
            if i is middleware:
                return True

        return False

    def run_polling(
        self,
        threads: int,
        *,
        delay_secs: Union[int, float] = 0,
        error_delay_secs: Union[int, float] = 5,
        limit: Optional[int] = None,
        timeout: Optional[int] = 10,
        allowed_updates: Optional[list[str]] = None
    ) -> None:
        if self._polling_is_used:
            raise DispatcherError("Polling cannot be run twice!")

        if self._server_is_used:
            raise DispatcherError("Polling cannot be run while the server is used!")

        if delay_secs < 0:
            raise ValueError("Delay seconds cannot be negative!")

        if error_delay_secs < 0:
            raise ValueError("Error delay seconds cannot be negative!")

        self._polling_is_used = True
        logger.debug("Polling is starting...")
        offset_update_id = None
        self._run_media_group_gathering_thread()
        self._run_thread_pool(threads)
        logger.info("Polling started.")

        try:
            while True:
                # noinspection PyBroadException
                try:
                    updates = self._bot.get_updates(
                        timeout_secs=timeout + 1 if timeout else None,
                        offset=offset_update_id,
                        limit=limit,
                        timeout=timeout,
                        allowed_updates=allowed_updates
                    )
                except Exception:
                    logger.exception("An error occurred while receiving updates!")
                    time.sleep(error_delay_secs)
                else:
                    for i in updates:
                        self._process_update(i)

                    if updates:
                        offset_update_id = updates[-1].update_id + 1

                    time.sleep(delay_secs)
        except KeyboardInterrupt:
            logger.info("Polling stopped.")
            self._finish_update_processing()
            self._media_group_gathering_thread = None
            self._thread_pool = None
            self._polling_is_used = False

    def run_server(
        self,
        threads: int,
        webhook_path: Optional[str] = None,
        *,
        host: str = "0.0.0.0",
        port: int = 443,
        ssl_certificate_path: Optional[str] = None,
        ssl_private_key_path: Optional[str] = None
    ) -> None:
        if self._server_is_used:
            raise DispatcherError("Server cannot be run twice!")

        if self._polling_is_used:
            raise DispatcherError("Server cannot be run while polling is used!")

        self._server_is_used = True
        logger.debug("Server is starting...")
        cherrypy.config.update({
            "server.socket_host": host,
            "server.socket_port": port,
            'environment': 'production',
            "engine.autoreload.on": False
        })

        if (ssl_certificate_path is not None) and (ssl_private_key_path is not None):
            cherrypy.config.update({
                "server.ssl_module": "builtin",
                "server.ssl_certificate": ssl_certificate_path,
                "server.ssl_private_key": ssl_private_key_path,
            })

        self._run_media_group_gathering_thread()
        self._run_thread_pool(threads)
        logger.info("Server started.")

        try:
            cherrypy.quickstart(
                root=ServerRoot(self._process_update),
                script_name=webhook_path.rstrip("/") if webhook_path else str()
            )
        finally:
            logger.info("Server stopped.")
            self._finish_update_processing()
            self._media_group_gathering_thread = None
            self._thread_pool = None
            self._server_is_used = False

    def drop_pending_updates(
        self,
        *,
        timeout_secs: Union[int, float, None] = None,
        with_delete_webhook: bool = True
    ) -> None:
        logger.debug("Dropping pending updates...")

        if with_delete_webhook:
            self._bot.delete_webhook(timeout_secs=timeout_secs, drop_pending_updates=True)
        else:
            updates = self._bot.get_updates(
                timeout_secs=timeout_secs,
                offset=-1
            )

            if updates:
                self._bot.get_updates(
                    timeout_secs=timeout_secs,
                    offset=updates[-1].update_id + 1
                )

        logger.info("Pending updates dropped.")

    def _add_event_handler(
        self,
        handler: AbstractEventHandler,
        event_type: EventType,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        filter_ = _get_event_filter(filter_)

        if not filter_.check_event_type(event_type):
            raise DispatcherError(f"{event_type!r} is not supported by this filter!")

        rate_limit = self._get_rate_limit(rate_limit)
        rate_limiter = RateLimiter(rate_limit) if rate_limit is not None else None
        self._event_handlers[event_type].append(
            EventHandler(
                handler=handler,
                filter=filter_,
                rate_limiter=rate_limiter
            )
        )

    def _check_event_handler(
        self,
        handler: AbstractEventHandler,
        event_type: EventType,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> bool:
        filter_ = _get_event_filter(filter_)
        rate_limit = self._get_rate_limit(rate_limit)

        for i in self._event_handlers[event_type]:
            if (
                (i.handler is handler)
                and (i.filter is filter_)
                and (i.rate_limit is rate_limit)
            ):
                return True

        return False

    def _get_event_handler(self, event: Event, event_type: EventType) -> Optional[EventHandler]:
        values = {}

        for i in self._event_handlers[event_type]:
            if i.filter.get_result(event, values):
                return i

    def _get_error_handler(self, error: Exception, event: Event) -> Optional[ErrorHandler]:
        values = {}

        for i in self._error_handlers:
            if i.filter.get_result(error, event, values):
                return i

    def _get_rate_limit(
        self,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> Optional[RateLimit]:
        return rate_limit if rate_limit is not NOT_SET else self._default_rate_limit

    def _run_media_group_gathering_thread(self) -> None:
        self._media_group_gathering_thread = Thread(
            target=self._run_media_group_gathering,
            daemon=True
        )
        self._media_group_gathering_thread.start()

    def _run_thread_pool(self, threads: int) -> None:
        self._thread_pool = ThreadPool(
            threads=threads,
            target=self._run_event_processing,
            args=(self._events,)
        )
        self._thread_pool.start()

    def _process_update(self, update: Update) -> None:
        logger.debug("Update received: %r.", update)

        try:
            event, content_type = update.content
        except UnknownUpdateContentError:
            logger.debug(
                "Update skipped because it contains an unknown content type: %r.",
                update
            )

            return

        event_type = EventType(content_type.value)

        if isinstance(event, Message):
            try:
                event.content  # noqa
            except UnknownMessageContentError:
                logger.debug(
                    "Event skipped because it contains an unknown content type: %r.",
                    event
                )

                return

            if event.media_group_id is not None:
                with self._media_group_message_lock:
                    if event.media_group_id in self._media_group_messages:
                        self._media_group_messages[event.media_group_id].add(event)
                    else:
                        self._media_group_messages[event.media_group_id] = TimedContainer(event)

                return

        if self._check_over_limit_event(event):
            logger.debug("Event from over limit chat dropped: %r.", event)
        else:
            logger.debug("Event added to queue: %r.", event)
            self._events.add_event(event, event_type)

    def _check_over_limit_event(self, event: Event) -> bool:
        return (
            self._drop_over_limit_events
            and get_event_chat_id(event) in self._bot.over_limit_chat_ids
        )

    def _finish_update_processing(self) -> None:
        logger.info("Finishing update processing...")
        self._events.wait_events()
        logger.info("Update processing finished.")

    def _run_media_group_gathering(self) -> NoReturn:
        while True:
            with self._media_group_message_lock:
                for i in tuple(self._media_group_messages):
                    if (time.monotonic() - self._media_group_messages[i].time) > 0.1:
                        event = MediaGroup(self._media_group_messages[i].items)
                        del self._media_group_messages[i]

                        if self._check_over_limit_event(event):
                            logger.debug("Event from over limit chat dropped: %r.", event)
                        else:
                            logger.debug("Event added to queue: %r.", event)
                            self._events.add_event(event, EventType.MEDIA_GROUP)

            time.sleep(_MEDIA_GROUP_GATHERING_DELAY_SECS)

    def _run_event_processing(self, events: EventQueue) -> NoReturn:
        while True:
            event = events.get_event()

            try:
                logger.debug("Event processing started: %r.", event)
                event_context.set(event.event)

                for i in self._middlewares:
                    i.pre_process_event(event.event, event.event_type)

                event_handler = self._get_event_handler(event.event, event.event_type)

                if event_handler is None:
                    logger.debug("No handler found for event: %r.", event)
                    continue

                event_handler_context.set(event_handler.handler)

                if (
                    (event_handler.rate_limiter is not None)
                    and event_handler.rate_limiter.process_call(event.chat_id, event.user_id)
                ):
                    continue

                for i in self._middlewares:
                    i.process_event(event.event, event.event_type)

                event_handler.handler.process_event(event.event)

                for i in self._middlewares:
                    i.post_process_event(event.event, event.event_type)
            except Exception as error:
                # noinspection PyBroadException
                try:
                    for i in self._middlewares:
                        i.pre_process_error(error, event.event, event.event_type)

                    error_handler = self._get_error_handler(error, event.event)

                    if error_handler is None:
                        raise

                    error_handler_context.set(error_handler.handler)

                    for i in self._middlewares:
                        i.process_error(error, event.event, event.event_type)

                    error_handler.handler.process_error(error, event.event)

                    for i in self._middlewares:
                        i.post_process_error(error, event.event, event.event_type)
                except Exception:
                    logger.exception("An error occurred while processing an event!")
            finally:
                events.set_event_as_processed(event)
                logger.debug("Event processing finished: %r.", event)


def _get_event_filter(
    filter_: Optional[AbstractEventBaseFilter] = None
) -> AbstractEventBaseFilter:
    return filter_ if filter_ is not None else _none_filter


def _get_error_filter(
    filter_: Optional[AbstractErrorBaseFilter] = None
) -> AbstractErrorBaseFilter:
    return filter_ if filter_ is not None else _error_none_filter



from typing import Optional

from telebox.dispatcher.utils.rate_limiter.rate_limit import RateLimit
from telebox.dispatcher.utils.rate_limiter.call_tracker import CallTracker


class RateLimiter:

    def __init__(self, limit: RateLimit):
        self.limit = limit
        self._call_tracker = CallTracker(
            chat_ttl_secs=limit.chat_limit_secs,
            user_ttl_secs=limit.user_limit_secs,
            chat_user_ttl_secs=limit.chat_user_limit_secs
        )
        self._processed_chat_ids: set[int] = set()
        self._processed_user_ids: set[int] = set()
        self._process_chat_user_ids: set[tuple[int, int]] = set()

    def process_call(self, chat_id: Optional[int], user_id: Optional[int]) -> bool:
        is_limited = False
        chat_id_is_not_none = chat_id is not None
        user_id_is_not_none = user_id is not None

        if chat_id_is_not_none and self.limit.check_chat_limit():
            chat_calls = self._call_tracker.get_chat_calls(chat_id)

            if chat_calls < self.limit.chat_limit_calls:
                self._processed_chat_ids.discard(chat_id)
            elif chat_calls == self.limit.chat_limit_calls:
                if chat_id not in self._processed_chat_ids:
                    self._processed_chat_ids.add(chat_id)
                    self.limit.process_chat_over_limit(chat_id)

                is_limited = True

        if user_id_is_not_none and self.limit.check_user_limit():
            user_calls = self._call_tracker.get_user_calls(user_id)

            if user_calls < self.limit.user_limit_calls:
                self._processed_user_ids.discard(user_id)
            elif user_calls == self.limit.user_limit_calls:
                if user_id not in self._processed_user_ids:
                    self._processed_user_ids.add(user_id)
                    self.limit.process_user_over_limit(user_id)

                is_limited = True

        if chat_id_is_not_none and user_id_is_not_none and self.limit.check_chat_user_limit():
            chat_user_calls = self._call_tracker.get_chat_user_calls(chat_id, user_id)
            chat_user_ids = (chat_id, user_id)

            if chat_user_calls < self.limit.chat_user_limit_calls:
                self._process_chat_user_ids.discard(chat_user_ids)
            elif chat_user_calls == self.limit.chat_user_limit_calls:
                if (chat_id, user_id) not in self._process_chat_user_ids:
                    self._process_chat_user_ids.add((chat_id, user_id))
                    self.limit.process_chat_user_over_limit(chat_id, user_id)

                is_limited = True

        if not is_limited:
            if chat_id_is_not_none:
                self._call_tracker.add_chat_call(chat_id)

            if user_id_is_not_none:
                self._call_tracker.add_user_call(user_id)

            if chat_id_is_not_none and user_id_is_not_none:
                self._call_tracker.add_chat_user_call(chat_id, user_id)

        return is_limited



from dataclasses import dataclass

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.types.types.inline_keyboard_button import InlineKeyboardButton


@dataclass(unsafe_hash=True)
class InlineKeyboardMarkup(Type):
    inline_keyboard: list[list[InlineKeyboardButton]]



from contextvars import ContextVar

from telebox.dispatcher.typing import Event
from telebox.dispatcher.handlers.event import AbstractEventHandler
from telebox.dispatcher.handlers.error import AbstractErrorHandler


event_context: ContextVar[Event] = ContextVar("event_context")
event_handler_context: ContextVar[AbstractEventHandler] = ContextVar("event_handler_context")
error_handler_context: ContextVar[AbstractErrorHandler] = ContextVar("error_handler_context")



from telebox.dispatcher.filters.event_filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.telegram_bot.types.types.message import Message


class SenderContactFilter(AbstractEventFilter):

    def get_event_types(self) -> set[EventType]:
        return {EventType.MESSAGE, EventType.CHANNEL_POST}

    def get_value(self, event: Message) -> bool:
        return (event.contact is not None) and (event.contact.user_id == event.user_id)

    def check_value(self, value: bool) -> bool:
        return value



from dataclasses import dataclass
from typing import Optional

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.types.types.user import User
from telebox.telegram_bot.types.types.location import Location


@dataclass(unsafe_hash=True)
class ChosenInlineResult(Type):
    result_id: str
    from_: User
    query: str
    location: Optional[Location] = None
    inline_message_id: Optional[str] = None

    @property
    def user_id(self) -> int:
        return self.from_.id



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.types.types.poll_option import PollOption
from telebox.telegram_bot.types.types.message_entity import MessageEntity


@dataclass(unsafe_hash=True)
class Poll(Type):
    id: str
    question: str
    options: list[PollOption]
    total_voter_count: int
    is_closed: bool
    is_anonymous: bool
    type: str
    allows_multiple_answers: bool
    correct_option_id: Optional[int] = None
    explanation: Optional[str] = None
    explanation_entities: Optional[list[MessageEntity]] = None
    open_period: Optional[int] = None
    close_date: Optional[datetime] = None



from dataclasses import dataclass
from typing import Optional

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.utils.users import get_full_name


@dataclass(unsafe_hash=True)
class InputContactMessageContent(Type):
    phone_number: str
    first_name: str
    last_name: Optional[str] = None
    vcard: Optional[str] = None

    @property
    def full_name(self) -> str:
        return get_full_name(self.first_name, self.last_name)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.types.types.user import User


@dataclass(unsafe_hash=True)
class ChatInviteLink(Type):
    invite_link: str
    creator: User
    creates_join_request: bool
    is_primary: bool
    is_revoked: bool
    name: Optional[str] = None
    expire_date: Optional[datetime] = None
    member_limit: Optional[int] = None
    pending_join_request_count: Optional[int] = None



from dataclasses import dataclass
from datetime import datetime

from telebox.telegram_bot.types.type import Type


@dataclass(unsafe_hash=True)
class PassportFile(Type):
    file_id: str
    file_unique_id: str
    file_size: int
    file_date: datetime



from typing import Union

import ujson

from telebox.utils.callback_data_builders.builder import AbstractCallbackDataBuilder


Key = Union[str, int]
Value = Union[str, int, float, bool, list, None]


class JSONCallbackDataBuilder(AbstractCallbackDataBuilder):

    def build(self, key: Key, value: Value = None) -> str:
        return ujson.dumps([key, value])

    def parse(self, data: str) -> tuple[Key, Value]:
        return tuple(ujson.loads(data))



from typing import Callable

from telebox.telegram_bot.utils.formatting import get_escaped_html_text, get_escaped_markdown_text


def get_html_text(template: str, /, *args, **kwargs) -> str:
    return _get_text(get_escaped_html_text, template, *args, **kwargs)


def get_markdown_text(template: str, /, *args, **kwargs) -> str:
    return _get_text(get_escaped_markdown_text, template, *args, **kwargs)


def _get_text(escaping_func: Callable[[str], str], template: str, /, *args, **kwargs) -> str:
    args = tuple(escaping_func(str(i)) for i in args)
    kwargs = {
        name: escaping_func(str(value))
        for name, value in kwargs.items()
    }

    return template.format(*args, **kwargs)



from dataclasses import dataclass
from typing import Optional

from telebox.telegram_bot.types.type import Type
from telebox.telegram_bot.types.types.user import User


@dataclass(unsafe_hash=True)
class MessageEntity(Type):
    type: str
    offset: int
    length: int
    url: Optional[str] = None
    user: Optional[User] = None
    language: Optional[str] = None
    custom_emoji_id: Optional[str] = None

    @property
    def end_offset(self) -> int:
        return self.offset + self.length



ABKHAZIAN = "ab"
AFAR = "aa"
AFRIKAANS = "af"
AKAN = "ak"
ALBANIAN = "sq"
AMHARIC = "am"
ARABIC = "ar"
ARAGONESE = "an"
ARMENIAN = "hy"
ASSAMESE = "as"
AVARIC = "av"
AVESTAN = "ae"
AYMARA = "ay"
AZERBAIJANI = "az"
BAMBARA = "bm"
BASHKIR = "ba"
BASQUE = "eu"
BELARUSIAN = "be"
BENGALI = "bn"
BISLAMA = "bi"
BOSNIAN = "bs"
BRETON = "br"
BULGARIAN = "bg"
BURMESE = "my"
CATALAN = "ca"
CHAMORRO = "ch"
CHECHEN = "ce"
CHICHEWA = "ny"
CHINESE = "zh"
CHURCH_SLAVIC = "cu"
CHUVASH = "cv"
CORNISH = "kw"
CORSICAN = "co"
CREE = "cr"
CROATIAN = "hr"
CZECH = "cs"
DANISH = "da"
DIVEHI = "dv"
DUTCH = "nl"
DZONGKHA = "dz"
ENGLISH = "en"
ESPERANTO = "eo"
ESTONIAN = "et"
EWE = "ee"
FAROESE = "fo"
FIJIAN = "fj"
FINNISH = "fi"
FRENCH = "fr"
WESTERN_FRISIAN = "fy"
FULAH = "ff"
GAELIC = "gd"
GALICIAN = "gl"
GANDA = "lg"
GEORGIAN = "ka"
GERMAN = "de"
GREEK = "el"
KALAALLISUT = "kl"
GUARANI = "gn"
GUJARATI = "gu"
HAITIAN = "ht"
HAUSA = "ha"
HEBREW = "he"
HERERO = "hz"
HINDI = "hi"
HIRI_MOTU = "ho"
HUNGARIAN = "hu"
ICELANDIC = "is"
IDO = "io"
IGBO = "ig"
INDONESIAN = "id"
INTERLINGUA = "ia"
INTERLINGUE = "ie"
INUKTITUT = "iu"
INUPIAQ = "ik"
IRISH = "ga"
ITALIAN = "it"
JAPANESE = "ja"
JAVANESE = "jv"
KANNADA = "kn"
KANURI = "kr"
KASHMIRI = "ks"
KAZAKH = "kk"
CENTRAL_KHMER = "km"
KIKUYU = "ki"
KINYARWANDA = "rw"
KIRGHIZ = "ky"
KOMI = "kv"
KONGO = "kg"
KOREAN = "ko"
KUANYAMA = "kj"
KURDISH = "ku"
LAO = "lo"
LATIN = "la"
LATVIAN = "lv"
LIMBURGAN = "li"
LINGALA = "ln"
LITHUANIAN = "lt"
LUBA_KATANGA = "lu"
LUXEMBOURGISH = "lb"
MACEDONIAN = "mk"
MALAGASY = "mg"
MALAY = "ms"
MALAYALAM = "ml"
MALTESE = "mt"
MANX = "gv"
MAORI = "mi"
MARATHI = "mr"
MARSHALLESE = "mh"
MONGOLIAN = "mn"
NAURU = "na"
NAVAJO = "nv"
NORTH_NDEBELE = "nd"
SOUTH_NDEBELE = "nr"
NDONGA = "ng"
NEPALI = "ne"
NORWEGIAN = "no"
NORWEGIAN_BOKMAL = "nb"
NORWEGIAN_NYNORSK = "nn"
SICHUAN_YI = "ii"
OCCITAN = "oc"
OJIBWA = "oj"
ORIYA = "or"
OROMO = "om"
OSSETIAN = "os"
PALI = "pi"
PASHTO = "ps"
PERSIAN = "fa"
POLISH = "pl"
PORTUGUESE = "pt"
PUNJABI = "pa"
QUECHUA = "qu"
ROMANIAN = "ro"
ROMANSH = "rm"
RUNDI = "rn"
RUSSIAN = "ru"
NORTHERN_SAMI = "se"
SAMOAN = "sm"
SANGO = "sg"
SANSKRIT = "sa"
SARDINIAN = "sc"
SERBIAN = "sr"
SHONA = "sn"
SINDHI = "sd"
SINHALA = "si"
SLOVAK = "sk"
SLOVENIAN = "sl"
SOMALI = "so"
SOUTHERN_SOTHO = "st"
SPANISH = "es"
SUNDANESE = "su"
SWAHILI = "sw"
SWATI = "ss"
SWEDISH = "sv"
TAGALOG = "tl"
TAHITIAN = "ty"
TAJIK = "tg"
TAMIL = "ta"
TATAR = "tt"
TELUGU = "te"
THAI = "th"
TIBETAN = "bo"
TIGRINYA = "ti"
TONGA = "to"
TSONGA = "ts"
TSWANA = "tn"
TURKISH = "tr"
TURKMEN = "tk"
TWI = "tw"
UIGHUR = "ug"
UKRAINIAN = "uk"
URDU = "ur"
UZBEK = "uz"
VENDA = "ve"
VIETNAMESE = "vi"
VOLAPUK = "vo"
WALLOON = "wa"
WELSH = "cy"
WOLOF = "wo"
XHOSA = "xh"
YIDDISH = "yi"
YORUBA = "yo"
ZHUANG = "za"
ZULU = "zu"



from typing import Iterable, Union

from telebox.utils.callback_data_builders.builder import AbstractCallbackDataBuilder


Key = str
Value = Union[str, Iterable[str], None]


class SeparatoryCallbackDataBuilder(AbstractCallbackDataBuilder):

    def __init__(self, separator: str):
        if not separator:
            raise ValueError(f"Invalid separator {separator!r}!")

        self._separator = separator

    def build(self, key: Key, value: Value = None) -> str:
        if self._separator in key:
            raise ValueError(f"Key {key!r} contains separator {self._separator!r}!")

        data = key

        if value is not None:
            if isinstance(value, str):
                value = [value]

            for i in value:
                if self._separator in i:
                    raise ValueError(f"Value {i!r} contains separator {self._separator!r}!")

                data += f"{self._separator}{i}"

        return data

    def parse(self, data: str) -> tuple[Key, Value]:
        key, *value = data.split(self._separator)

        if not value:
            value = None
        elif len(value) == 1:
            value = value[0]

        return key, value



from .telegram_bot import ContextTelegramBot
from .state_machine import ContextStateMachine
from .vars import event_context, event_handler_context, error_handler_context


__all__ = [
    "ContextTelegramBot",
    "ContextStateMachine",
    "event_context",
    "event_handler_context",
    "error_handler_context"
]



from abc import ABC, abstractmethod
from typing import Optional, Any


class AbstractErrorBaseFilter(ABC):

    def __invert__(self):
        return InvertErrorFilter(self)

    def __and__(self, other):
        if isinstance(other, ConjunctionErrorFilter):
            return ConjunctionErrorFilter(self, *other)
        elif isinstance(other, AbstractErrorBaseFilter):
            return ConjunctionErrorFilter(self, other)

        return NotImplemented

    def __or__(self, other):
        if isinstance(other, DisjunctionErrorFilter):
            return DisjunctionErrorFilter(self, *other)
        elif isinstance(other, AbstractErrorBaseFilter):
            return DisjunctionErrorFilter(self, other)

        return NotImplemented

    @abstractmethod
    def get_result(self, error, event, values: Optional[dict[type, Any]] = None) -> bool:
        pass


class AbstractErrorFilter(AbstractErrorBaseFilter, ABC):

    @abstractmethod
    def get_value(self, error, event):
        pass

    @abstractmethod
    def check_value(self, value) -> bool:
        pass

    def get_result(self, error, event, values: Optional[dict[type, Any]] = None) -> bool:
        values = values or {}

        try:
            value = values[type(self)]
        except KeyError:
            value = values[type(self)] = self.get_value(error, event)

        return self.check_value(value)


class InvertErrorFilter(AbstractErrorBaseFilter):

    def __init__(self, filter_: AbstractErrorBaseFilter):
        self.filter = filter_

    def __repr__(self):
        return f"{type(self).__name__}({self.filter!r})"

    def __invert__(self):
        return self.filter

    def get_result(self, error, event, values: Optional[dict[type, Any]] = None) -> bool:
        return not self.filter.get_result(error, event, values)


class ConjunctionErrorFilter(AbstractErrorBaseFilter):

    def __init__(self, *filters: AbstractErrorBaseFilter):
        self.filters = filters

    def __repr__(self):
        return f"{type(self).__name__}({', '.join(repr(i) for i in self.filters)})"

    def __iter__(self):
        return iter(self.filters)

    def __and__(self, other):
        if isinstance(other, ConjunctionErrorFilter):
            return ConjunctionErrorFilter(*self, *other)
        elif isinstance(other, AbstractErrorBaseFilter):
            return ConjunctionErrorFilter(*self, other)

        return NotImplemented

    def get_result(self, error, event, values: Optional[dict[type, Any]] = None) -> bool:
        return all(i.get_result(error, event, values) for i in self.filters)


class DisjunctionErrorFilter(AbstractErrorBaseFilter):

    def __init__(self, *filters: AbstractErrorBaseFilter):
        self.filters = filters

    def __repr__(self):
        return f"{type(self).__name__}({', '.join(repr(i) for i in self.filters)})"

    def __iter__(self):
        return iter(self.filters)

    def __or__(self, other):
        if isinstance(other, DisjunctionErrorFilter):
            return DisjunctionErrorFilter(*self, *other)
        elif isinstance(other, AbstractErrorBaseFilter):
            return DisjunctionErrorFilter(*self, other)

        return NotImplemented

    def get_result(self, error, event, values: Optional[dict[type, Any]] = None) -> bool:
        return any(i.get_result(error, event, values) for i in self.filters)
